/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * @module nmodule/webChart/rc/fe/StartEndTimeRangeEditor
 * @private
 */
define([
    'baja!',
    'jquery',
    'Promise',
    'nmodule/webEditors/rc/fe/baja/ComplexCompositeEditor',
    'nmodule/webChart/rc/webChartUtil',
    'nmodule/webEditors/rc/fe/fe',
    'bajaux/events',
    'bajaux/Widget',
    'moment',
    'hbs!nmodule/webChart/rc/fe/startEndTimeRangeEditorStructure',
    'nmodule/webEditors/rc/fe/baja/AbsTimeRangeEditor',
    'nmodule/webEditors/rc/fe/baseEditorSet',
    'baja!' +
      'baja:AbsTime,' +
      'baja:AbsTimeRange,' +
      'baja:TimeRange,' +
      'baja:Time,' +
      'baja:Weekday,' +
      'bql:DynamicTimeRangeType',
    'css!nmodule/webChart/rc/ChartWidgetStyle',
  ],
  function(baja,
           $,
           Promise,
           ComplexCompositeEditor,
           webChartUtil,
           fe,
           events,
           Widget,
           moment,
           startEndTimeRangeEditorStructure) {
    "use strict";

    /**
     * A StartEndTimeRangeEditor is the editor for webChart:WebChartTimeRangEditor.
     *
     * @private
     * @class
     * @alias module:nmodule/webChart/rc/fe/StartEndTimeRangeEditor
     * @extends module:nmodule/webEditors/fe/baja/ComplexCompositeEditor
     */

    var StartEndTimeRangeEditor = function(params) {
      var that = this;
      ComplexCompositeEditor.apply(this, arguments);

      that.validators().add(function (diff) {
        var value = that.value().newCopy();
        return diff.apply(value).then(startBeforeEnd);
      });
    };
    StartEndTimeRangeEditor.prototype = Object.create(ComplexCompositeEditor.prototype);
    StartEndTimeRangeEditor.prototype.constructor = StartEndTimeRangeEditor;

    /**
     * Return all the properties of the WebChartTimeRange
     * @returns {string[]}
     */
    StartEndTimeRangeEditor.prototype.getSlotFilter = function() {
      return ['startFixed', 'startTime', 'endFixed', 'endTime'];
    };

    StartEndTimeRangeEditor.prototype.$buildSubEditorDom = function(key) {
      var jq = this.jq(),
          dom = $('.js-' + key, jq);
      return dom;
    };

    var MODIFY_EVENT = events.MODIFY_EVENT;

    /**
     * initialize the dom and register all event handlers
     * @param {jQuery} dom
     */
    StartEndTimeRangeEditor.prototype.doInitialize = function(dom) {
      var that = this;

      dom.on(MODIFY_EVENT, '.slot-startFixed, .slot-endFixed', function() {
        that.checkTimeEditors();
        that.setModified(true); //ensure that re-validation occurs when these slots change
      });

      var div = $("<div></div>");
      div.html(startEndTimeRangeEditorStructure());
      that.jq().append(div);
      //doing this last allows squelching of the premature modify event
      return ComplexCompositeEditor.prototype.doInitialize.apply(this, arguments);
    };

    /**
     * Load the current value and setup any dependant slot interactions
     * @returns {Promise|*}
     * @param value {baja.Value}
     */
    StartEndTimeRangeEditor.prototype.doLoad = function(value) {
      var that = this;

      return Promise.resolve(ComplexCompositeEditor.prototype.doLoad.apply(this, arguments))
        .then(function() {
          return that.checkTimeEditors();
        });
    };

    /**
     * Ensure that the value is valid time range
     * @param {webChart:WebChartTimeRange} value
     * @inner
     */
    function startBeforeEnd(value) {
      var startTime = value.get('startTime'),
          endTime = value.get('endTime'),
          startFixed = value.get('startFixed'),
          endFixed = value.get('endFixed');
      if(startFixed && endFixed && startTime.getJsDate().getTime() > endTime.getJsDate().getTime()){
        //TODO: if we ever start showing the validation errors to the user, we should lexicon this
        throw new Error('start time not before end time');
      }
    }

    /**
     * Return the Start Time element.
     * @private
     * @returns {jquery}
     */
    StartEndTimeRangeEditor.prototype.$getStartTimeEditor = function() {
      return this.jq().find('.slot-startTime').data('widget');
    };

    /**
     * Return the End Time  element.
     * @private
     * @returns {jquery}
     */
    StartEndTimeRangeEditor.prototype.$getEndTimeEditor = function() {
      return this.jq().find('.slot-endTime').data('widget');
    };

    /**
     * Return the Start Fixed element.
     * @private
     * @returns {jquery}
     */
    StartEndTimeRangeEditor.prototype.$getStartFixedEditor = function() {
      return this.jq().find('.slot-startFixed').data('widget');
    };

    /**
     * Return the End Fixed element.
     * @private
     * @returns {jquery}
     */
    StartEndTimeRangeEditor.prototype.$getEndFixedEditor = function() {
      return this.jq().find('.slot-endFixed').data('widget');
    };

    /**
     * Setup any interdependency between the checkboxes and the other editors
     * @returns {Promise}
     */
    StartEndTimeRangeEditor.prototype.checkTimeEditors = function() {
      var that = this,
          startTimeEditor = that.$getStartTimeEditor(),
          endTimeEditor = that.$getEndTimeEditor(),
          startFixedEditor = that.$getStartFixedEditor(),
          endFixedEditor = that.$getEndFixedEditor();

      return Promise.join(startFixedEditor.read(), endFixedEditor.read())
        .spread(function(startEnabled, endEnabled) {
          return Promise.join(startTimeEditor.setEnabled(startEnabled),
            endTimeEditor.setEnabled(endEnabled));
        });
    };

    /**
     * register this editor automatically if the file is included
     * @inner
     */
    function register() {
      fe.register('webChart:WebChartTimeRange',
        'nmodule/webChart/rc/fe/StartEndTimeRangeEditor');
    }

    register();

    return StartEndTimeRangeEditor;
  });
