/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson and JJ Frankovich
 */

/*jshint browser: true*/

/**
 * A Layer in a Line Chart to handle the X-Axis.
 *
 * @module nmodule/webChart/rc/line/XAxisLayer
 * @private
 */
define(["d3",
  "nmodule/webChart/rc/webChartUtil",
  "nmodule/webChart/rc/localeUtil",
  "nmodule/webChart/rc/model/modelUtil",
  "nmodule/webChart/rc/model/samplingUtil",
  "moment"], function (d3, webChartUtil, localeUtil, modelUtil, samplingUtil, moment) {
  "use strict";

  var XAxisLayer = function(graph) {
    var that = this;

    that.$graph = graph;
  };

  XAxisLayer.className = function(d) {
    return d ? 'x-axis-minor' : 'x-axis-major';
  };

  /**
   * Provide a fallback Time for domain calculation. Times are slightly broadened if bar is use to show the full bar
   *
   * @param graph
   * @param propertyName
   * @param {moment} fallbackValue, domain value to use if time unknown
   * @returns {moment} the domain value to use
   * @inner
   */
  function fixedTime(graph, seriesList, propertyName, fallbackValue) {
    var timeRange = graph.widget().model().timeRange();
    if((timeRange.getPeriod().getOrdinal() > 0 && graph.widget().isTimeZoom()) ||
      (graph.widget().isDataZoom() && !graph.widget().model().hasPoints() && timeRange.getPeriod().getOrdinal() !== 0)){

      if(timeRange.getPeriod().getOrdinal() === 1){  //timeRange
        return timeRange.get(propertyName).getJsDate();
      }else{
        var results = webChartUtil.getStartAndEndDateFromPeriod(timeRange.getPeriod());
        if(propertyName === "startTime"){
          return results.start;
        }else if(propertyName === "endTime"){
          return results.end;
        }
      }
    }else if(!graph.data().length ||
      (!graph.widget().model().hasPoints() && timeRange.getPeriod().getOrdinal() === 0)){
      return fallbackValue;
    }
  }

  /**
   * Broaden the times if bar (to show the full bar period) or shape (to show extra point)
   * @param {Line} graph
   * @param {Array.<BaseSeries>} seriesList
   * @param {Array.<Number>} domain
   * @returns {Array.<Number>} adjusted Domain if bar is present
   */
  XAxisLayer.broadenDomain = function(graph, seriesList, domain) {
    var newDomain = [domain[0], domain[1]];
    if (modelUtil.hasBar(seriesList) && graph.widget().model().samplingMillis()) {
      var amount = graph.widget().model().samplingMillis() / 2;

      if (amount) {
        newDomain[0] = modelUtil.safeMomentAdd(newDomain[0], -amount);
        newDomain[1] = modelUtil.safeMomentAdd(newDomain[1], amount);
      }
    }

    if (modelUtil.hasVisibleShade(seriesList)) {
      newDomain[1] = new Date(Math.max(moment(newDomain[1]).toDate(),
        modelUtil.safeMomentAdd(domain[1], modelUtil.getTimeForPixels(graph.widget().model(), 6)).toDate()));
    }

    return newDomain;
  };

  /**
   * Setup domain based on the present information, data will usually be used, but fallbackDomaincan be utilized if the data doesn't provide it.

   * @param {module:nmodule/webChart/rc/line/Line} graph
   * @param {Array.<module:nmodule/webChart/rc/model/BaseSeries> | module:nmodule/webChart/rc/model/BaseSeries} seriesList or series for domain calculation
   * @param {Array.<moment>} fallbackDomain
   * @returns {Array.<moment>} adjusted Domain
   * @inner
   */
  XAxisLayer.domain = function (graph,
                                seriesList,
                                fallbackDomain
                                ) {
    if (!(seriesList instanceof Array)) {
      seriesList = [ seriesList ];
    }

    var minFixed = fixedTime(graph, seriesList, "startTime", fallbackDomain[0]);
    var maxFixed = fixedTime(graph, seriesList, "endTime", fallbackDomain[1]);
    if (minFixed !== undefined && maxFixed !== undefined) {
      return [ minFixed, maxFixed ];
    }

    var min = Number.POSITIVE_INFINITY,
      max = Number.NEGATIVE_INFINITY;

    for (var i = 0; i < seriesList.length; i++) {
      var points = seriesList[i].points();
      if (points && points[0] !== undefined) {
        min = Math.min(min, points[0].x.getTime());
      }
      if (points && points[points.length - 1] !== undefined) {
        max = Math.max(max, points[points.length - 1].x.getTime());
      }
    }

    if (min === Number.POSITIVE_INFINITY && max === Number.NEGATIVE_INFINITY) {
      return fallbackDomain;
    }

    return [ moment(min), moment(max) ];
  };

  /**
   * Ensure a domain is in range
   * @param fallbackDomain
   * @param desiredDomain
   * @return {{tooHigh: boolean, tooLow: boolean, tooSmall: boolean, newDomain: *}}
   */
  XAxisLayer.checkDomain = function(desiredDomain, fallbackDomain) {
    var results = {
          tooHigh: false,
          tooLow: false,
          tooSmall: false,
          newDomain: desiredDomain
        },
        x1 = desiredDomain[0],
        x2 = desiredDomain[1],
        d;

    if(x1 instanceof Date){
      if(x1.getFullYear() < 1900 || x2.getFullYear() <= 1900){
        results.tooLow = true;
        d = new Date();
        d.setFullYear(1900);
        results.newDomain[0] = d;
        if(x2.getFullYear() <= 1900){
          d = new Date();
          d.setFullYear(1901);
          results.newDomain[1] = d;
        }
      }
      if(x1.getFullYear() > 2200 || x2.getFullYear() >= 2200){
        results.tooHigh = true;
        d = new Date();
        d.setFullYear(2200);
        results.newDomain[1] = moment(d).endOf("year").toDate();

        if(x1.getFullYear() >= 2200){
          d = new Date();
          d.setFullYear(2199);
          results.newDomain[1] = moment(d).endOf("year").toDate();
        }

      }else if(moment(x2).diff(moment(x1)) < 5000){ //time difference is too small (5 seconds)
        results.tooSmall = true;
        results.newDomain = fallbackDomain;
      }
    }

    if(!results.newDomain[0]){
      results.newDomain[0] = fallbackDomain[0];
    }

    if(!results.newDomain[1]){
      results.newDomain[1] = fallbackDomain[1];
    }
    return results;
  };

  XAxisLayer.prototype.getPeriodDisplay = function() {
    var that = this,
        xdomain = that.$scale.domain(),
        start = moment(xdomain[0]),
        end = moment(xdomain[1]),
        dateFormat = localeUtil.getD3DateFormat(),
        timeFormat = localeUtil.getD3TimeFormat(),
        hourMinuteFormat = localeUtil.getD3HourMinuteFormat(),
        startDisplay = localeUtil.format(dateFormat, start),
        endDisplay = localeUtil.format(dateFormat, end);

    if(startDisplay === "Invalid date" || endDisplay === "Invalid date"){
      return "";
    }
    else if(startDisplay === endDisplay){
      if(localeUtil.format(hourMinuteFormat, start) !== localeUtil.format(hourMinuteFormat, end)){ //same day, but at least a different minute
        if(start.format("h") === end.format("h") && localeUtil.hasPeriod()){ //am and pm is same to add to time summary
          return startDisplay + " " + start.format("a");
        }else{
          return startDisplay;
        }
      }
      else{ //minute (and second) display is required on high zoom level
        return startDisplay + " (" + webChartUtil.lex.get("dateTo", localeUtil.format(timeFormat, start), localeUtil.format(timeFormat, end)) + ")";
      }
    }
    else{
      return webChartUtil.lex.get("dateTo", startDisplay, endDisplay);
    }
  };

  /**
   * Get the primary data set for the x axis, pick the first series with data to zoom in on
   * @returns {webChart.BaseSeries}
   */
  XAxisLayer.prototype.getPrimarySeries = function() {
    var that = this,
        graph = that.$graph,
        seriesList = graph.data(),
        i;

    for(i = 0; i < seriesList.length; i++){
      if(seriesList[i].points().length){
        return seriesList[i];
      }
    }

    for(i = 0; i < seriesList.length; i++){
      if(!seriesList[i].points().length){
        return seriesList[i];
      }
    }
  };

  /**
   * Initialize the XAxisLayer to cache the models timeScale.
   */
  XAxisLayer.prototype.initialize = function() {
    var that = this,
        graph = that.$graph;
    that.$scale = graph.widget().model().timeScale().scale();
  };

  XAxisLayer.prototype.redraw = function() {
    var that = this,
        graph = that.$graph,
        chart = graph.chartd3(),
        widget = graph.widget(),
        settings = widget.settings(),
        ticksGroup = chart.selectAll('g.ticks'),
        dom = graph.widget().jq3(),
        seriesList,
        gx,
        gxe,
        desiredDomain,
        fallbackDomain,
        scale = that.$scale,
        fx = function(d) {
          return scale.tickFormat(10)(d);
        },
        tx = function(d) {
          return 'translate(' + scale(d) + ',0)';
        };

    // Reset the scales to map to the new dimensions of the container
    scale.range([0, widget.dimensions().width]);

    //now that focus data is available, calculate the sampling data for rest of draw

    if (widget.isDataZoom() || widget.isTimeZoom()) {

      if(settings.getDataZoomScope() === 'primary'){
        seriesList = that.getPrimarySeries();
        if(seriesList){
          seriesList = [seriesList]; //ensure its an array if present
        }
      }else{
        seriesList = graph.data();
      }

      fallbackDomain = [ moment().startOf("day"), moment().endOf("day") ];
      desiredDomain = XAxisLayer.domain(graph, seriesList, fallbackDomain);
      scale.domain(XAxisLayer.checkDomain(desiredDomain, fallbackDomain).newDomain);
    }

    widget.model().startSampling();

    function xaxisDown() {
      graph.widget().manualZoom();
      var p = d3.mouse(chart.node());
      that.$xAxisDragData = scale.invert(p[0]);
    }

    function xaxisDrag() {
      var p,
          newDomain,
          dragDistance,
          checkDomainResults,
          rupx,
          xaxis1, xaxis2, xextent;

      if(typeof that.$xAxisDragData === "object"){
        p = d3.mouse(chart.node());
        graph.widget().manualZoom();

        dom.style('cursor', 'w-resize');

        rupx = +scale.invert(p[0]);
        xaxis1 = +scale.domain()[0];
        xaxis2 = +scale.domain()[1];

        xextent = xaxis2 - xaxis1;

        if(rupx !== 0){
          dragDistance = that.$xAxisDragData - rupx;

          //newDomain = [new Date(xaxis1), new Date(xaxis1 + xextent + dragDistance)]; //zoom into future
          newDomain = [new Date(xaxis1 + dragDistance), new Date(xaxis2)];//zoom into past
          checkDomainResults = XAxisLayer.checkDomain(newDomain, scale.domain());
          if(checkDomainResults.tooSmall || checkDomainResults.tooHigh || checkDomainResults.tooLow){
            newDomain = checkDomainResults.newDomain;
          }
          scale.domain(newDomain);
          graph.widget().redrawSoon();
        }
        d3.event.preventDefault();
        d3.event.stopPropagation();
      }
    }

    function xaxisUp() {
      dom.style('cursor', 'auto');

      if(typeof that.$xAxisDragData === "object"){
        graph.widget().redrawSoon();
        that.$xAxisDragData = undefined;
        d3.event.preventDefault();
        d3.event.stopPropagation();
      }
    }


    var tickCount = Math.floor(widget.dimensions().width/80);

    // Regenerate x-ticks…
    gx = ticksGroup.selectAll('g.x')
      .data(scale.ticks(tickCount))
      .attr('transform', tx);

    gx.select('text')
      .text(fx);

    gxe = gx.enter().insert('g', 'a')
      .attr('class', 'x')
      .attr('transform', tx);

    if(settings.hasShowGrid()){
      gxe.append('line')
        .attr('y1', 0)
        .attr('class', XAxisLayer.className);
    }

    gxe.append('text')
      .attr('class', 'axis x-axis')
      .attr('dy', '1.5em')
      .attr('text-anchor', 'middle')
      .text(fx);

    dom.on('mousemove.xaxisdrag', xaxisDrag)
      .on('touchmove.xaxisdrag', xaxisDrag)
      .on('mouseup.xaxisdrag', xaxisUp)
      .on('touchend.xaxisdrag', xaxisUp);

    var lines = ticksGroup
      .selectAll('g.x line');

    if(settings.hasShowGrid()){

      if(!lines.node()){ //need to add lines back
        ticksGroup
          .selectAll('g.x')
          .append('line')
          .attr('y1', 0)
          .attr('class', XAxisLayer.className)
          .attr('y2', widget.dimensions().height);
      }else{
        lines.attr('y2', widget.dimensions().height);
      }
    }else{
      lines.remove();
    }


    ticksGroup
      .selectAll('g.x text')
      .attr('y', widget.dimensions().height);


    chart
      .selectAll('rect.x-ticks')
      .attr('x', -20)
      .attr('y', widget.dimensions().height)
      .attr('height', "1.75em")
      .attr('width', widget.dimensions().width + 25)
      .on('mousedown.xaxisdrag', xaxisDown)
      .on('touchstart.xaxisdrag', xaxisDown);

    gx.exit().remove();

    widget.model().addExtraPoints();

    if ((widget.isDataZoom() || widget.isTimeZoom())){
      var  existingDomain = scale.domain(),
           newDomain = XAxisLayer.broadenDomain(graph, seriesList, existingDomain);

      scale.domain(newDomain);
    }
  };

  XAxisLayer.prototype.getScale = function() {
    return this.$scale;
  };

  XAxisLayer.prototype.destroy = function() {
    var dom = this.$graph.widget().jq3();
    if(dom){
      dom
        .on('mousemove.xaxisdrag', null)
        .on('touchmove.xaxisdrag', null)
        .on('mouseup.xaxisdrag', null)
        .on('touchend.xaxisdrag', null);
    }
  };



  return XAxisLayer;
});
