/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Danesh Kamal
 */

/**
 * @module nmodule/alarm/rc/util/alarmUtils
 * @private
 */
define(['baja!',
        'baja!alarm:AlarmAlgorithm,alarm:AlarmSourceExt,' +
          'control:ControlPoint',
        'lex!alarm',
        'nmodule/webEditors/rc/util/textUtils',
        'Promise'], function (
         baja,
         types,
         lexs,
         textUtils,
         Promise) {

  'use strict';

  var alarmLex = lexs[0];

  /**
   * Utility functions for alarm editors and views.
   *
   * @private
   * @exports nmodule/alarm/rc/util/alarmUtils
   */
  var exports = {};

  /**
   * Get the parent `AlarmSourceExt` of an `AlarmAlgorithm`.
   *
   * @param {baja.Component} algorithm an `alarm:AlarmAlgorithm` component
   * @returns {baja.Component|null} the parent `alarm:AlarmSourceExt` component,
   * or `null` if not found
   * @see BAlarmAlgorithm#getParentExt
   */
  exports.getParentExt = function (algorithm) {
    var parent = baja.hasType(algorithm, 'alarm:AlarmAlgorithm') &&
          algorithm.getParent();
    if (!baja.hasType(parent, 'alarm:AlarmSourceExt')) {
      return null;
    }
    return parent;
  };

  /**
   * Get the grandparent `control:ControlPoint` of an `alarm:AlarmAlgorithm`
   * component.
   *
   * @param {baja.Component} algorithm an `alarm:AlarmAlgorithm` component
   * @returns {baja.Component|null} the grandparent `control:ControlPoint`
   * component, or `null` if not found
   * @see BAlarmAlgorithm#getParentPoint
   */
  exports.getParentPoint = function (algorithm) {
    var ext = exports.getParentExt(algorithm),
        point = ext && ext.getParent();
    if (!baja.hasType(point, 'control:ControlPoint')) {
      return null;
    }
    return point;
  };

  /**
   * Get the facets of an `alarm:AlarmAlgorithm`'s containing control point.
   *
   * @param {baja.Component} algorithm an `alarm:AlarmAlgorithm` component
   * @returns {Promise} promise to be resolved with the `facets` slot of the
   * containing control point, or with `baja.Facets.DEFAULT` if not found
   * @see BAlarmAlgorithm#getPointFacets
   */
  exports.getPointFacets = function (algorithm) {
    var point = exports.getParentPoint(algorithm);
    if (point) {
      return Promise.resolve(point.isMounted() && point.lease())
        .then(function () { return point.get('facets'); });
    }
    return Promise.resolve(baja.Facets.DEFAULT);
  };

  /**
   * Returns the display name for the alarm data column.
   *
   * If a display name can't be found in a lexicon, the key will
   * attempt to be converted to a friendly name.
   *
   * @since Niagara 4.6
   *
   * @param  {String} key The key for the display name.
   * @returns {String} The display name.
   */
  exports.getAlarmDataDisplayName = function (key) {
    return alarmLex.get('alarmData.' + key) || textUtils.toFriendly(key);
  };

  return exports;
});
